package net.simon987.server;

import net.simon987.server.event.GameEvent;
import net.simon987.server.event.GameEventDispatcher;
import net.simon987.server.event.TickEvent;
import net.simon987.server.game.GameUniverse;
import net.simon987.server.game.World;
import net.simon987.server.logging.LogManager;
import net.simon987.server.plugin.PluginManager;
import net.simon987.server.plugin.ServerPlugin;
import net.simon987.server.user.User;
import net.simon987.server.webserver.SocketServer;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

public class GameServer implements Runnable {

	public final static GameServer INSTANCE = new GameServer();

	private GameUniverse gameUniverse;
	private GameEventDispatcher eventDispatcher;
	private PluginManager pluginManager;

	private ServerConfiguration config;

	private SocketServer socketServer;

	private int maxExecutionTime;

	public ArrayList<byte[]> saveArchive;
	
	public int maxArchiveSize;

	public GameServer() {

		this.config = new ServerConfiguration(new File("config.properties"));

		gameUniverse = new GameUniverse(config);
		pluginManager = new PluginManager();

		maxExecutionTime = config.getInt("user_timeout");

		// Load all plugins in plugins folder, if it doesn't exist, create it
		File pluginDir = new File("plugins/");
		File[] pluginDirListing = pluginDir.listFiles();

		if (pluginDirListing != null) {
			for (File pluginFile : pluginDirListing) {

				if (pluginFile.getName().endsWith(".jar")) {
					pluginManager.load(pluginFile);
				}

			}
		} else {
			if (!pluginDir.mkdir()) {
				LogManager.LOGGER.severe("Couldn't create plugin directory");
			}
		}

		eventDispatcher = new GameEventDispatcher(pluginManager);
		
		saveArchive = new ArrayList<byte[]>();
		
		maxArchiveSize = config.getInt("max_archive_size");
	}

	public GameUniverse getGameUniverse() {
		return gameUniverse;
	}

	public GameEventDispatcher getEventDispatcher() {
		return eventDispatcher;
	}

	@Override
	public void run() {
		LogManager.LOGGER.info("(G) Started game loop");

		long startTime; // Start time of the loop
		long uTime; // update time
		long waitTime; // time to wait

		boolean running = true;

		while (running) {

			startTime = System.currentTimeMillis();

			tick();

			uTime = System.currentTimeMillis() - startTime;
			waitTime = config.getInt("tick_length") - uTime;

			LogManager.LOGGER.info("Wait time : " + waitTime + "ms | Update time: " + uTime + "ms | "
					+ (int) (((double) uTime / waitTime) * 100) + "% load");

			try {
				if (waitTime >= 0) {
					Thread.sleep(waitTime);
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}

		}

	}

	private void tick() {
		gameUniverse.incrementTime();

		// Dispatch tick event
		GameEvent event = new TickEvent(gameUniverse.getTime());
		GameServer.INSTANCE.getEventDispatcher().dispatch(event); // Ignore cancellation

		// Process user code
		ArrayList<User> users_ = gameUniverse.getUsers();
		for (User user : users_) {

			if (user.getCpu() != null) {
				try {

					int timeout = Math.min(user.getControlledUnit().getEnergy(), maxExecutionTime);

					user.getCpu().reset();
					int cost = user.getCpu().execute(timeout);
					user.getControlledUnit().spendEnergy(cost);

				} catch (Exception e) {
					LogManager.LOGGER.severe("Error executing " + user.getUsername() + "'s code");
					e.printStackTrace();
				}

			}
		}

		// Process each worlds
		// Avoid concurrent modification
		ArrayList<World> worlds = new ArrayList<>(gameUniverse.getWorlds());
		for (World world : worlds) {
			world.update();
		}

		// Save
		if (gameUniverse.getTime() % config.getInt("save_interval") == 0) {
			save(new File("save.json"));
		}

		socketServer.tick();

		LogManager.LOGGER.info("Processed " + gameUniverse.getWorlds().size() + " worlds");
	}

	/**
	 * Save game universe to file in JSON format
	 *
	 * @param file
	 *            JSON file to save
	 */
	public void save(File file) {

		if (new File(new File("save.json").getAbsolutePath()).exists()) {
			saveArchive.add(ZipUtils.bytifyFile("save.json"));
			while(saveArchive.size() > maxArchiveSize) {
				saveArchive.remove(0);
			}
		}
		
		try {
			FileWriter fileWriter = new FileWriter(file);

			JSONObject universe = gameUniverse.serialise();

			JSONArray plugins = new JSONArray();

			for (ServerPlugin plugin : pluginManager.getPlugins()) {
				plugins.add(plugin.serialise());
			}

			universe.put("plugins", plugins);

			fileWriter.write(universe.toJSONString());
			fileWriter.close();

			LogManager.LOGGER.info("Saved to file " + file.getName());

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	public ServerConfiguration getConfig() {
		return config;
	}

	public PluginManager getPluginManager() {
		return pluginManager;
	}

	public void setSocketServer(SocketServer socketServer) {
		this.socketServer = socketServer;
	}
	
	public ArrayList<byte[]> getSaveArchive() {
		return this.saveArchive;
	}
}
